#include <windows.h>
#include <stdio.h>
#include <time.h>
#include "GLES20Emu.h"
#include "OpenGL30Drv.h"
#include "Vector.h"
#include "Transform.h"
#include "Shader.h"
#include "Tga.h"

static int n_width = 640;
static int n_height = 480;
// window size constants

static const bool b_fullscreen = false;
static const bool b_centered = true;
static const bool b_topmost = false;
// window placement constants

static const bool b_window_frame = true;
static const bool b_show_cursor = true;
// window style constants

static const char *p_s_window_name = "OpenGL ES 2.0 window";
static const char *p_s_class_name = "my_ogles_wnd_class";
// window name constants

/**
 *	@def __GL_WINDOW_TRACKBALL_STYLE_CAMERA
 *	@brief if defined, camera can be controlled using a mouse in trackball style motion;
 *		otherwise, camera can be controled using a mouse and WSAD keys in FPS style motion
 */
#define __GL_WINDOW_TRACKBALL_STYLE_CAMERA

static Vector3f v_camera_pos(0, 0, -2.5f);
static float f_angle_x = 30 * f_pi / 180, f_angle_y = 10 * f_pi / 180, f_angle_z = 0;
// camera

#include "Windowing.h"

/**
 *	@brief vraci "pointer" do VBO
 *	@param[in] off je offset v bytech, na ktery ma vraceny pointer ukazovat
 */
#define p_OffsetInVBO(off) ((void*)(off))

GLuint m_n_vertex_buffer, m_n_index_buffer, m_n_vertex_array_object;
// OpenGL buffery

class CConsole {
protected:
	int m_n_width, m_n_height;
	int m_n_cursor_pos; // position of the cursor on the last line

	char *m_p_s_text; // the buffer for console text is allocated staticaly
	struct TLine {
		char *p_s_text;
		size_t n_length; // why do i even need this?
	};
	TLine *m_p_line_list; // lines are stored in several buffers which need to be concatenated before use
	int m_n_last_line; // round buffer

public:
	CConsole(int n_width, int n_height)
		:m_n_width(n_width), m_n_height(n_height), m_n_cursor_pos(0),
		m_p_s_text(new(std::nothrow) char[(n_width + 1) * n_height]),
		m_p_line_list(new(std::nothrow) TLine[n_height])
	{
		if(m_p_s_text && m_p_line_list) {
			for(int i = 0; i < n_height; ++ i) {
				m_p_line_list[i].p_s_text = m_p_s_text + (n_width + 1) * i;
				*m_p_line_list[i].p_s_text = 0;
				m_p_line_list[i].n_length = 0;
			}
		}
	}

	~CConsole()
	{
		if(m_p_s_text)
			delete[] m_p_s_text;
		if(m_p_line_list)
			delete[] m_p_line_list;
	}

	bool b_Status() const
	{
		return m_p_s_text != 0 && m_p_line_list != 0;
	}

	bool Print(const char *p_s_text)
	{
		if(!p_s_text)
			return Print("(null)");
		// debug

		for(; *p_s_text; ++ p_s_text) {
			if(*p_s_text == '\n') {
			} else if(*p_s_text == '\t') {
				int n_next_cursor_pos = (m_n_cursor_pos + 4) & ~3;
				// calculate the next cursor position (for tab size 8 use "+ 4) & ~7".
				// note this only works for powers of two)
			} else if(*p_s_text == '\b') {
				if(m_n_cursor_pos) {
					-- m_n_cursor_pos;
					m_p_line_list[m_n_last_line].p_s_text[m_n_cursor_pos] = 0;
					-- m_p_line_list[m_n_last_line].n_length;
				}
			} else if(*p_s_text == '\r') {
				m_n_cursor_pos = 0;
			} else {
			}

			;
		}
	}

protected:
	CConsole(const CConsole &UNUSED(r_console)) {} // don't copy objects of this class; use pointers instead
	CConsole & operator =(const CConsole &UNUSED(r_console)) { return *this; } // don't copy objects of this class; use pointers instead
};

void PrepareBuffersToRenderABox()
{
	const float p_box_vertices[] = {
		-1, -1, -1,								//      7-------6
		+1, -1, -1,								//     /|      /|
		+1, -1, +1,								//    / |     / | y+
		-1, -1, +1,								//   4-------5  | ^ z+
		-1, +1, -1,								//   |  3----|--2 |/
		+1, +1, -1,								//   | /     | /  +-> x+
		+1, +1, +1,								//   |/      |/
		-1, +1, +1,								//   0-------1
	};

	const unsigned short p_box_indices[] = {
		//1, 0, 2, 3, 7, 0, 4, 1, 5, 2, 6, 7, 5, 4 // this works, but it generates CCW triangles. now what?
		0, 1, 3, 2, 6, 1, 5, 0, 4, 3, 7, 6, 4, 5 // this works (left-to-right flip).
	};
	//     inwards-looking     |     outwards-looking
	//                         |
	//  4---5                  |                  4---5
	//  |  /|                  |                  |\  |
	//  |/  |                  |                  |  \|
	//  7---6---5---4---7      |      6---5---4---7---6
	//   \  |  /|  /|  /|\     |     /|\  |\  |\  |  /
	//     \|/  |/  |/  |  \   |   /  |  \|  \|  \|/
	//      2---1---0---3---2  |  3---2---1---0---3
	//                  |  /|  |  |\  |
	//                  |/  |  |  |  \|
	//                  0---1  |  1---0
	// a nice inwards-looking box (GL_TRIANGLE_STRIP, 14 vertices)

	glGenBuffers(1, &m_n_vertex_buffer);
	glBindBuffer(GL_ARRAY_BUFFER, m_n_vertex_buffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(p_box_vertices), p_box_vertices, GL_STATIC_DRAW);

	glGenBuffers(1, &m_n_index_buffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_n_index_buffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(p_box_indices), p_box_indices, GL_STATIC_DRAW);

	if(GLEH_ARB_VERTEX_ARRAY_OBJECT) {
		glGenVertexArrays(1, &m_n_vertex_array_object);
		glBindVertexArray(m_n_vertex_array_object);
		// create and bind terrain VAO to remember the buffer configuration below

		glBindBuffer(GL_ARRAY_BUFFER, m_n_vertex_buffer);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, false, 3 * sizeof(float), 0);
		// set vertex coordinate attribute

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_n_index_buffer);
		// set vertex indices buffer

		glBindVertexArray(0);
		// unbind VAO
	}
}

void PrepareBuffersToRenderATexturedBox()
{
	const float p_box_vertices[] = {
		0, 1, -1, -1, 1, 1, 1, 1, -1, 1, 1, 0, 1, 1, 1, 0, 0, -1, 1, 1,
		1, 1, -1, -1, -1, 1, 0, -1, 1, -1, 0, 0, 1, 1, -1, 0, 1, 1, -1, -1,
		0, 0, -1, 1, -1, 0, 1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, -1,
		1, 0, -1, -1, -1, 0, 0, 1, -1, -1, 0, 1, 1, -1, 1, 1, 1, -1, -1, 1,
		1, 1, 1, -1, -1, 1, 0, 1, 1, -1, 0, 0, 1, 1, 1, 0, 1, 1, -1, 1,
		0, 1, -1, -1, -1, 1, 1, -1, -1, 1, 1, 0, -1, 1, 1, 0, 0, -1, 1, -1};
	// vrcholy krychle, vzdy 2 souradnice textury, pak 3 souradnice vrcholu; vrcholy jsou usporadane poporade po ctvericich, tvoricich polygony jednotlivych sten

	const unsigned short p_box_indices[] = {	//  0     1
		0, 1, 2, 0, 2, 3,					//  *-----*
		4, 5, 6, 4, 6, 7,					//  |\    |
		8, 9, 10, 8, 10, 11,				//  |  \  |
		12, 13, 14, 12, 14, 15,				//  |    \|
		16, 17, 18, 16, 18, 19,				//  *-----*
		20, 21, 22, 20, 22, 23};			//  3     2
	// indexy trojuhelniku krychle (kazda stena se sklada ze dvou trojuhelniku, viz. ascii art); v OpenGL 3.0 uz nemame GL_QUADS a tak musime takto kreslit trojuhelniky

	glGenBuffers(1, &m_n_vertex_buffer);
	glBindBuffer(GL_ARRAY_BUFFER, m_n_vertex_buffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(p_box_vertices), p_box_vertices, GL_STATIC_DRAW);
	// vyrobime buffer objekt, do ktereho zkopirujeme geometrii naseho modelu (do jednoho bufferu ulozime vsechny souradnice, tzn. texturovaci / normaly / barvy / pozice, atp. jeden buffer muze obsahovat data pro jeden nebo i vice modelu)

	glGenBuffers(1, &m_n_index_buffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_n_index_buffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(p_box_indices), p_box_indices, GL_STATIC_DRAW);
	// vytvorime druhy buffer objekt, do ktereho zkopirujeme indexy vrcholu naseho modelu (jedna se o jiny typ bufferu)

	if(GLEH_ARB_VERTEX_ARRAY_OBJECT) {
		glGenVertexArrays(1, &m_n_vertex_array_object);
		glBindVertexArray(m_n_vertex_array_object);
		// create and bind terrain VAO to remember the buffer configuration below

		glBindBuffer(GL_ARRAY_BUFFER, m_n_vertex_buffer);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 2, GL_FLOAT, false, 5 * sizeof(float), p_OffsetInVBO(0));
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, false, 5 * sizeof(float), p_OffsetInVBO(2 * sizeof(float)));
		// set vertex coordinate attribute

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_n_index_buffer);
		// set vertex indices buffer

		glBindVertexArray(0);
		// unbind VAO
	}
}

void RenderBox()
{
	if(GLEH_ARB_VERTEX_ARRAY_OBJECT)
		glBindVertexArray(m_n_vertex_array_object);
	else {
		glBindBuffer(GL_ARRAY_BUFFER, m_n_vertex_buffer);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, false, 3 * sizeof(float), 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_n_index_buffer);
	}

	glDrawElements(GL_TRIANGLE_STRIP, 14, GL_UNSIGNED_SHORT, 0);
}

void RenderTexturedBox()
{
	if(GLEH_ARB_VERTEX_ARRAY_OBJECT)
		glBindVertexArray(m_n_vertex_array_object);
	else {
		glBindBuffer(GL_ARRAY_BUFFER, m_n_vertex_buffer);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 2, GL_FLOAT, false, 5 * sizeof(float), p_OffsetInVBO(0));
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, false, 5 * sizeof(float), p_OffsetInVBO(2 * sizeof(float)));
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_n_index_buffer);
	}

	glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_SHORT, 0);
}

struct TMySimpleShader : public TShader {
	GLint n_modelview_matrix_uniform;

	bool Compile()
	{
		const char *p_s_vertex_shader =
			"precision highp float;\n" // specifikace presnosti
			"\n"
			"attribute vec3 v_pos;\n" // atributy - vstup z dat vrcholu
			"\n"
			"uniform mat4 t_modelview_projection_matrix;\n" // parametr shaderu - transformacni matice
			"\n"
			"varying vec3 v_color;\n" // vystup - barva vrcholu
			"\n"
			"void main()\n"
			"{\n"
			"    gl_Position = t_modelview_projection_matrix * vec4(v_pos, 1.0);\n" // musime zapsat pozici
			"    v_color = v_pos * .5 + .5;\n" // spocitame barvu vrcholu pro fragment shader
			"}\n";
		const char *p_s_fragment_shader =
			"precision highp float;\n" // specifikace presnosti
			"\n"
			"varying vec3 v_color;\n" // vstup z vertex shaderu
			"\n"
			"void main()\n"
			"{\n"
			"    gl_FragColor = vec4(v_color, 1.0);\n" // zapise vystupni barvu
			"}\n";
		// zdrojove kody pro vertex / fragment shader (OpenGL 3 uz nepouziva specifikator 'varying', jinak je kod stejny)

		if(!CGLShaderCompiler::CompileProgram(*this, p_s_vertex_shader, p_s_fragment_shader,
		   "my shader", "v_pos: 1;", true, true, true))
			return false;
		// just call the compiler

		n_modelview_matrix_uniform = glGetUniformLocation(n_program_object, "t_modelview_projection_matrix");
		// nacte adresy uniformu v pameti GPU

		return true;
	}

	void Bind(float *p_modelview_projection_matrix)
	{
		TShader::Bind();
		// napred nabindovat shader!

		glUniformMatrix4fv(n_modelview_matrix_uniform, 1, GL_FALSE, p_modelview_projection_matrix);
		// set the projection matrix
	}
};

struct TMyComplexShader : public TShader {
	GLint n_modelview_matrix_uniform;
	GLint n_texture_unit_uniform;

	bool Compile()
	{
		const char *p_s_vertex_shader =
			"precision highp float;\n" // specifikace presnosti
			"\n"
			"attribute vec2 v_tex;\n"
			"attribute vec3 v_pos;\n" // atributy - vstup z dat vrcholu
			"\n"
			"uniform mat4 t_modelview_projection_matrix;\n" // parametr shaderu - transformacni matice
			"\n"
			"varying vec2 v_texcoord;\n"
			"varying vec3 v_normal;\n" // vystup - souradnice textury
			"\n"
			"void main()\n"
			"{\n"
			"    gl_Position = t_modelview_projection_matrix * vec4(v_pos, 1.0);\n" // musime zapsat pozici
			"    v_normal = normalize(v_pos);\n" // je to krabice, normala je pozice
			"    v_texcoord = v_tex;\n" // zkopirujeme souradnice textury pro fragment shader
			"}\n";
		const char *p_s_fragment_shader =
			"precision highp float;\n" // specifikace presnosti
			"\n"
			"varying vec2 v_texcoord;\n"
			"varying vec2 v_normal;\n" // vstupy z vertex shaderu
			"\n"
			"uniform sampler2D n_texture_unit;\n" // cislo texturovaci jednotky
			"\n"
			"void main()\n"
			"{\n"
			"    vec4 v_tex_color = texture2D(n_texture_unit, v_texcoord);\n" // precte texturu
			"    gl_FragColor = v_tex_color;\n" // zapise vystupni barvu
			"}\n";
		// zdrojove kody pro vertex / fragment shader (OpenGL 3 uz nepouziva specifikator 'varying', jinak je kod stejny)

		if(!CGLShaderCompiler::CompileProgram(*this, p_s_vertex_shader, p_s_fragment_shader,
		   "my shader", "v_pos: 1; v_tex: 0;", true, true, true))
			return false;
		// just call the compiler

		n_modelview_matrix_uniform = glGetUniformLocation(n_program_object, "t_modelview_projection_matrix");
		n_texture_unit_uniform = glGetUniformLocation(n_program_object, "n_texture_unit");
		// nacte adresy uniformu v pameti GPU

		return true;
	}

	void Bind(float *p_modelview_projection_matrix)
	{
		TShader::Bind();
		// napred nabindovat shader!

		glUniformMatrix4fv(n_modelview_matrix_uniform, 1, GL_FALSE, p_modelview_projection_matrix);
		// set the projection matrix

		glUniform1i(n_texture_unit_uniform, 0);
		// rekne ze textura je v texturovaci jednotce 0
	}
};

struct TMyAnimatedShader : public TShader {
	GLint n_modelview_matrix_uniform;
	GLint n_texture_unit_uniform0;
	GLint n_texture_unit_uniform1;
	GLint n_time_uniform;

	bool Compile()
	{
		const char *p_s_vertex_shader =
			"precision highp float;\n" // specifikace presnosti
			"\n"
			"attribute vec2 v_tex;\n"
			"attribute vec3 v_pos;\n" // atributy - vstup z dat vrcholu
			"\n"
			"uniform float f_time;\n"
			"uniform mat4 t_modelview_projection_matrix;\n" // parametr shaderu - transformacni matice
			"\n"
			"varying vec2 v_texcoord;\n"
			"varying vec3 v_normal;\n" // vystup - souradnice textury
			"\n"
			"void main()\n"
			"{\n"
			"    gl_Position = t_modelview_projection_matrix * vec4(v_pos + vec3(0, sin(f_time * 5.0) * .2, 0), 1.0);\n" // musime zapsat pozici
			"    v_normal = normalize(v_pos);\n" // je to krabice, normala je pozice
			"    v_texcoord = v_tex;\n" // zkopirujeme souradnice textury pro fragment shader
			"}\n";
		const char *p_s_fragment_shader =
			"precision highp float;\n" // specifikace presnosti
			"\n"
			"varying vec2 v_texcoord;\n"
			"varying vec2 v_normal;\n" // vstupy z vertex shaderu
			"\n"
			"uniform float f_time;\n"
			"uniform sampler2D n_texture_unit0, n_texture_unit1;\n" // cislo texturovaci jednotky
			"\n"
			"void main()\n"
			"{\n"
			"    vec4 v_tex_color = texture2D(n_texture_unit0, v_texcoord);\n" // precte texturu
			"    vec4 v_fire_color = texture2D(n_texture_unit1, v_texcoord + vec2(sin(f_time * .75), f_time)) +\n"
			"        texture2D(n_texture_unit1, v_texcoord + vec2(cos(f_time + 1.0), f_time * .75));\n" // precte texturu
			"    gl_FragColor = mix(v_fire_color, v_tex_color, v_tex_color.a);\n" // zapise vystupni barvu
			"}\n";
		// zdrojove kody pro vertex / fragment shader (OpenGL 3 uz nepouziva specifikator 'varying', jinak je kod stejny)

		if(!CGLShaderCompiler::CompileProgram(*this, p_s_vertex_shader, p_s_fragment_shader,
		   "my shader", "v_pos: 1; v_tex: 0;", true, true, true))
			return false;
		// just call the compiler

		n_modelview_matrix_uniform = glGetUniformLocation(n_program_object, "t_modelview_projection_matrix");
		n_texture_unit_uniform0 = glGetUniformLocation(n_program_object, "n_texture_unit0");
		n_texture_unit_uniform1 = glGetUniformLocation(n_program_object, "n_texture_unit1");
		n_time_uniform = glGetUniformLocation(n_program_object, "f_time");
		// nacte adresy uniformu v pameti GPU

		return true;
	}

	void Bind(float *p_modelview_projection_matrix, float f_time)
	{
		TShader::Bind();
		// napred nabindovat shader!

		glUniformMatrix4fv(n_modelview_matrix_uniform, 1, GL_FALSE, p_modelview_projection_matrix);
		// set the projection matrix

		glUniform1f(n_time_uniform, f_time);
		// nastavi cas pro animaci

		glUniform1i(n_texture_unit_uniform0, 0);
		glUniform1i(n_texture_unit_uniform1, 1);
		// rekne ze textury je v texturovaci jednotce 0, 1
	}
};

TMyAnimatedShader t_shader;

bool CompileShader()
{
	return t_shader.Compile();
}

/**
 *	@brief vyrobi OpenGL texturu z bitmapy
 *	@param[in] p_bitmap je obrazek textury
 *	@param[in] n_internal_format je pozadovany format ulozeni textury na graficke karte (format obrazku je v typu TBmp vzdy GL_RGBA)
 *	@return vraci id textury (neuvazuje neuspech)
 */
GLuint n_GLTextureFromBitmap(const TBmp *p_bitmap, GLenum n_internal_format = GL_RGBA)
{
	GLuint n_tex;
	glGenTextures(1, &n_tex);
	glBindTexture(GL_TEXTURE_2D, n_tex);
	//glEnable(GL_TEXTURE_2D); !! OpenGL 3 uz nema enable / disable textur, shader si sam cte textury ktere potrebuje !!
	// vyrobi a nastavi novou 2D texturu, zatim nema zadne data

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// nastavime filtrovani

	glTexImage2D(GL_TEXTURE_2D, 0, n_internal_format, p_bitmap->n_width, p_bitmap->n_height,
		0, GL_RGBA, GL_UNSIGNED_BYTE, p_bitmap->p_buffer);
	// specifikujeme data textury

	glGenerateMipmap(GL_TEXTURE_2D);
	// nechame OpenGL spocitat jeji mipmapy

	return n_tex;
}

GLuint n_GLTextureFromFile(const char *p_s_filename, GLenum n_internal_format = GL_RGBA)
{
	TBmp *p_texture = CTgaCodec::p_Load_TGA(p_s_filename);
	if(!p_texture)
		return 0;
	GLint n_texture = n_GLTextureFromBitmap(p_texture);
	p_texture->Delete();
	// nahraje texturu

	return n_texture;
}

GLuint n_box_texture, n_fire_texture;

/**
 *	@brief load application's resources here, put cleanup code inside onCleanup()
 *	@return Returns true on success, false on failure (exits the application).
 */
bool onInitialize()
{
	PrepareBuffersToRenderATexturedBox(); // prepare buffers for a box

	n_box_texture = n_GLTextureFromFile("box.tga");
	n_fire_texture = n_GLTextureFromFile("fire.tga");

	return CompileShader(); // compile a shader
}

/**
 *	@brief called when the window is about to close
 *		(OpenGL is still running, h_wnd is still valid)
 */
void onCleanup()
{
	// todo - add cleanup code here
}

/**
 *	@brief called when the window is resized
 *
 *	@param[in] n_x, n_y, n_new_width and n_new_height are coordinates of
 *		client rectangle of the window
 *
 *	usualy glViewport is called here, n_width and n_height should
 *	be set to n_new_width and n_new_height, respectively
 */
void onResize(int n_x, int n_y, int n_new_width, int n_new_height)
{
	glViewport(n_x, n_y, n_width = n_new_width, n_height = n_new_height);
	// set new viewport
}

static Vector2f v_old_gl_mouse(0, 0);

/**
 *	@brief called when one of mouse buttons is pressed / released
 *		- n_button is one of button_Left, button_Middle or button_Right
 *		- b_pressing_button is true in case the button is being pressed,
 *		  or false in case it's being released
 *		- n_x, n_y are window-system mouse coordinates
 */
void onMouseClick(int n_button, bool b_pressing_button, int n_x, int n_y)
{
	Vector2f v_gl_coord(float(n_x) / n_width * 2 - 1, float(n_y) / n_height * 2 - 1);
	// calculate normalized mouse coordinates

	switch(n_button) {
	case button_Left:
		break;
	case button_Middle:
		break;
	case button_Right:
		break;
	}
	// handle different button codes

	v_old_gl_mouse = v_gl_coord;
}

/**
 *	@brief called when moving mouse, while one (or more) mouse buttons are pressed
 *		- n_button is combination of button_Left, button_Middle and button_Right or-ed together
 *		- n_x, n_y are window-system mouse coordinates
 */
void onMouseMotion(int n_button, int n_x, int n_y)
{
	Vector2f v_gl_coord(float(n_x) / n_width * 2 - 1, float(n_y) / n_height * 2 - 1);
	// calculate normalized mouse coordinates

	if(n_button & button_Left) {
		f_angle_y += v_gl_coord.x - v_old_gl_mouse.x;
#ifdef __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		f_angle_x -= v_gl_coord.y - v_old_gl_mouse.y;
#else // __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		f_angle_x += v_gl_coord.y - v_old_gl_mouse.y;
#endif // __GL_WINDOW_TRACKBALL_STYLE_CAMERA
	}
	if(n_button & button_Middle) {
#ifdef __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		f_angle_z += v_gl_coord.x - v_old_gl_mouse.x;
#endif // __GL_WINDOW_TRACKBALL_STYLE_CAMERA
	}
	if(n_button & button_Right) {
#ifdef __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		v_camera_pos.z += v_gl_coord.x - v_old_gl_mouse.x;
#else // __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		f_angle_z += v_gl_coord.x - v_old_gl_mouse.x;
#endif // __GL_WINDOW_TRACKBALL_STYLE_CAMERA
	}

	v_old_gl_mouse = v_gl_coord;
}

/**
 *	@brief called when moving mouse, while no mouse buttons are pressed
 *		- n_x, n_y are window-system mouse coordinates
 */
void onPassiveMouseMotion(int n_x, int n_y)
{
	Vector2f v_gl_coord(float(n_x) / n_width * 2 - 1, float(n_y) / n_height * 2 - 1);
	// calculate normalized mouse coordinates
}

void onKeyPressed(int n_key)
{
	if(n_key == VK_ESCAPE) {
		Exit();
	} else {
#ifndef __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		const float k = .5f;
		switch(n_key) {
		case 'W':
			v_camera_pos += Vector3f(-k * float(sin(f_angle_y)) * float(cos(f_angle_x)),
				k * float(sin(f_angle_x)), float(cos(f_angle_y)) * k * float(cos(f_angle_x)));
			break;
		case 'S':
			v_camera_pos -= Vector3f(-k * float(sin(f_angle_y)) * float(cos(f_angle_x)),
				k * float(sin(f_angle_x)), float(cos(f_angle_y)) * k * float(cos(f_angle_x)));
			break;
		case 'A':
			v_camera_pos += Vector3f(k * float(cos(f_angle_y)), 0, k * float(sin(f_angle_y)));
			break;
		case 'D':
			v_camera_pos -= Vector3f(k * float(cos(f_angle_y)), 0, k * float(sin(f_angle_y)));
			break;
		}
		glPostRedisplay();
#endif // !__GL_WINDOW_TRACKBALL_STYLE_CAMERA
	}
}

/**
 *	@brief called when WM_PAINT message arrives
 *		- can use glSwapBuffers() or glPostRedisplay()
 */
void onDisplay()
{
	Matrix4f t_mvp;
	{
		Matrix4f t_projection;
		CGLTransform::Perspective(t_projection, 90, float(n_width) / n_height, .01f, 1000);
		// spocitame matici perspektivni projekce

		Matrix4f t_camera_matrix;
		t_camera_matrix.Identity();
#ifdef __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		t_camera_matrix.Translate(v_camera_pos);
#endif // __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		t_camera_matrix.RotateZ(f_angle_z);
		t_camera_matrix.RotateX(f_angle_x);
		t_camera_matrix.RotateY(f_angle_y);
#ifndef __GL_WINDOW_TRACKBALL_STYLE_CAMERA
		t_camera_matrix.Translate(v_camera_pos);
#endif // !__GL_WINDOW_TRACKBALL_STYLE_CAMERA
		// calculate final camera matrix (with pitch and roll)

		Matrix4f t_model_matrix;
		t_model_matrix.Identity();
		// calculate object matrix

		Matrix4f t_modelview = t_camera_matrix * t_model_matrix;
		// fuse object and camera matrix to modelview

		t_mvp = t_projection * t_modelview;
	}
	// spocitame modelview - projection matici, kterou potrebujeme k transformaci vrcholu

	glClearColor(.5f, .5f, .5f, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);

	t_shader.Bind(&t_mvp[0][0], float(clock()) / CLOCKS_PER_SEC);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, n_box_texture);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, n_fire_texture);

	RenderTexturedBox();

	// todo - fill in your paint code here

	glSwapBuffers();
	if(glGetError() != GL_NO_ERROR)
		SetWindowText(h_wnd, "OpenGL error(s) occured!");

	glPostRedisplay();
}
