/*
								+---------------------------------+
								|                                 |
								|  ***   OpenGL 3.0 driver   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2009  |
								|                                 |
								|         OpenGL30Drv.cpp         |
								|                                 |
								+---------------------------------+
*/

/**
 *	@file OpenGL30Drv.cpp
 *	@author -tHE SWINe-
 *	@brief (Windows) OpenGL 3.0 driver
 */

#include <stdio.h>
#include "OpenGL30Drv.h"

/*
 *								=== WGL_ARB_create_context / GLX_ARB_create_context ===
 */

#ifndef WGL_ARB_create_context
#define WGL_ARB_create_context 1
#define __GENERATE_WGL_ARB_create_context__

#if defined(_WIN32) || defined(_WIN64)
extern HGLRC (GLApi *wglCreateContextAttribsARB)(HDC, HGLRC, const int*);
#else //_WIN32 || _WIN64
extern GLXContext (GLApi *glXCreateContextAttribsARB)(Display*, GLXFBConfig, GLXContext, Bool, const int *);
//extern GLXFBConfig *(GLApi *glXChooseFBConfigARB)(Display*, int, const int*, int*);
//extern XVisualInfo *(GLApi *glXGetVisualFromFBConfigARB)(Display*, GLXFBConfig);
#define glXChooseFBConfigARB glXChooseFBConfig
#define glXGetVisualFromFBConfigARB glXGetVisualFromFBConfig
#endif //_WIN32 || _WIN64

#define WGL_CONTEXT_MAJOR_VERSION_ARB				0x2091
#define WGL_CONTEXT_MINOR_VERSION_ARB				0x2092
#define WGL_CONTEXT_LAYER_PLANE_ARB					0x2093
#define WGL_CONTEXT_FLAGS_ARB						0x2094
#define WGL_CONTEXT_PROFILE_MASK_ARB				0x9126
#define WGL_CONTEXT_DEBUG_BIT_ARB					0x0001
#define WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB		0x0002

#define WGL_CONTEXT_CORE_PROFILE_BIT_ARB			0x00000001
#define WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB	0x00000002

#define GL_ERROR_INVALID_VERSION_ARB				0x2095
#define GL_ERROR_INVALID_PROFILE_ARB				0x2096

#endif //WGL_ARB_create_context

/*
 *								=== ~WGL_ARB_create_context / GLX_ARB_create_context ===
 */

/*
 *								=== CGL30Driver ===
 */

#ifdef __GENERATE_WGL_ARB_create_context__

#if defined(_WIN32) || defined(_WIN64)
HGLRC (GLApi *wglCreateContextAttribsARB)(HDC, HGLRC, const int*) = 0;
#else //_WIN32 || _WIN64
GLXContext (GLApi *glXCreateContextAttribsARB)(Display*, GLXFBConfig, GLXContext, Bool, const int *) = 0;
//GLXFBConfig *(GLApi *glXChooseFBConfigARB)(Display*, int, const int*, int*) = 0;
//XVisualInfo *(GLApi *glXGetVisualFromFBConfigARB)(Display*, GLXFBConfig) = 0;
#endif //_WIN32 || _WIN64

#endif //__GENERATE_WGL_ARB_create_context__

int CGL30Driver::n_GetCreateContextARBFuncPointers()
{
	int n_failed_functions = 0;

#ifdef __GENERATE_WGL_ARB_create_context__

#if defined(_WIN32) || defined(_WIN64)
	if(!(wglCreateContextAttribsARB = (HGLRC(GLApi*)(HDC, HGLRC, const int*))wglGetProcAddress("wglCreateContextAttribsARB"))) ++ n_failed_functions;
#else //_WIN32 || _WIN64
	if(!(glXCreateContextAttribsARB = (GLXContext(GLApi*)(Display*, GLXFBConfig, GLXContext, Bool, const int*))glXGetProcAddress((const GLubyte*)"glXCreateContextAttribsARB"))) ++ n_failed_functions;
	//if(!(glXChooseFBConfigARB = (GLXFBConfig*(GLApi*)(Display*, int, const int*, int*))glXGetProcAddressARB((const GLubyte*)"glXChooseFBConfigARB"))) ++ n_failed_functions;
  	//if(!(glXGetVisualFromFBConfigARB = (XVisualInfo*(GLApi*)(Display*, GLXFBConfig))glXGetProcAddressARB((const GLubyte*)"glXGetVisualFromFBConfigARB"))) ++ n_failed_functions;
#endif //_WIN32 || _WIN64

#endif //__GENERATE_WGL_ARB_create_context__

	return n_failed_functions;
}

CGL30Driver::CGL30Driver()
	:m_b_status(false)
{
}

CGL30Driver::~CGL30Driver()
{
    if(m_b_status)
        Shutdown();
}

#if defined(_WIN32) || defined(_WIN64)
bool CGL30Driver::Init(HWND h_wnd, bool b_forward_compatible,
	int n_opengl_major, int n_opengl_minor, int n_width, int n_height, int n_bpp,
	int n_depth_bpp, int n_stencil_bpp, bool b_fullscreen)
#else //_WIN32 || _WIN64
bool CGL30Driver::Init(Window h_wnd, Display *h_display, bool b_forward_compatible,
	int n_opengl_major, int n_opengl_minor, int n_width, int n_height, int n_bpp,
	int n_depth_bpp, int n_stencil_bpp, bool b_fullscreen)
#endif //_WIN32 || _WIN64
{
	if(m_b_status)
		Shutdown();

#if defined(_WIN32) || defined(_WIN64)
    if(b_fullscreen) {
	    DEVMODE t_device_mode;
        memset(&t_device_mode, 0, sizeof(t_device_mode));
        t_device_mode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
        t_device_mode.dmSize = sizeof(DEVMODE);
        t_device_mode.dmPelsWidth = n_width;
        t_device_mode.dmPelsHeight = n_height;
        t_device_mode.dmBitsPerPel = n_bpp;
        // sets device mode for fullscreen mode

        if(ChangeDisplaySettings(&t_device_mode, 4) != DISP_CHANGE_SUCCESSFUL) {
            if(MessageBoxA(NULL, "Unable to set fullscreen mode.\n"
               "Use windowed mode instead?", "Fullscreen", MB_YESNO | MB_ICONEXCLAMATION) == IDYES)
                m_b_fullscreen = 0;
            else
                return false;
        }
        // tries to change the mode, allows fallback to windowed

        ShowCursor(false);
    }

    memset(&m_t_pixel_format, 0, sizeof(PIXELFORMATDESCRIPTOR));
    m_t_pixel_format.nSize = sizeof(PIXELFORMATDESCRIPTOR);
    m_t_pixel_format.nVersion = 1;
    m_t_pixel_format.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER | PFD_SWAP_EXCHANGE;
    m_t_pixel_format.dwLayerMask = PFD_MAIN_PLANE;
    m_t_pixel_format.iPixelType = PFD_TYPE_RGBA;
    m_t_pixel_format.cColorBits = n_bpp;
    m_t_pixel_format.cDepthBits = n_depth_bpp;
    m_t_pixel_format.cAccumBits = 0;
    m_t_pixel_format.cStencilBits = n_stencil_bpp;
    // sets pixel format descriptor

    if(!(m_h_dc = GetDC(m_h_wnd = h_wnd))) {
        MessageBoxA(NULL, "Unable to create a GL device context.", "OpenGL driver", MB_OK | MB_ICONERROR);
        return false;
    }
    if(!(m_n_pixel_format_id = ChoosePixelFormat(m_h_dc, &m_t_pixel_format))) {
        MessageBoxA(NULL, "Unable to find a suitable pixelformat.", "OpenGL driver", MB_OK | MB_ICONERROR);
        return false;
    }
    if(!SetPixelFormat(m_h_dc, m_n_pixel_format_id, &m_t_pixel_format)) {
        MessageBoxA(NULL, "Unable to set the pixelformat.", "OpenGL driver", MB_OK | MB_ICONERROR);
        return false;
    }

	HGLRC h_gl_rc;
	if(!(h_gl_rc = wglCreateContext(m_h_dc))) {
        MessageBoxA(NULL, "Unable to create a GL rendering context.", "OpenGL driver", MB_OK | MB_ICONERROR);
        return false;
	}
    if(!wglMakeCurrent(m_h_dc, h_gl_rc)) {
        MessageBoxA(NULL, "Unable to activate the GL rendering context.", "OpenGL driver", MB_OK | MB_ICONERROR);
        return false;
    }
	// create and activate dummy context

	if(b_forward_compatible) {
		if(!wglCreateContextAttribsARB && n_GetCreateContextARBFuncPointers()) {
			MessageBoxA(NULL, "OpenGL 3.0 not supported by window system\n"
				"(WGL_ARB_create_context not present).", "OpenGL driver", MB_OK | MB_ICONERROR);
			return false;
		}
		// needs wglCreateContextAttribsARB extension

		const int p_params[] = {
			WGL_CONTEXT_LAYER_PLANE_ARB, 0, // main plane
			WGL_CONTEXT_MAJOR_VERSION_ARB, n_opengl_major,
			WGL_CONTEXT_MINOR_VERSION_ARB, n_opengl_minor,
			WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB, // want forward compatible context
			//WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB, // want core profile // nvidia drivers have problem with this
			0
		};
		if(!(m_h_glrc = wglCreateContextAttribsARB(m_h_dc, 0, p_params))) {
			//MessageBoxA(NULL, "Unable to create OpenGL 3.0 rendering context.", "OpenGL driver", MB_OK | MB_ICONERROR);
			return false;
		}
		if(!wglMakeCurrent(m_h_dc, m_h_glrc)) {
			MessageBoxA(NULL, "Unable to activate the GL rendering context.", "OpenGL driver", MB_OK | MB_ICONERROR);
			return false;
		}
		// creates forward compatible OpenGL context

		wglDeleteContext(h_gl_rc);
		// deletes the dummy context
	} else
		m_h_glrc = h_gl_rc; // otherwise just use the dummy context
#else //_WIN32 || _WIN64
	n_GetCreateContextARBFuncPointers();
	/*if(!glXChooseFBConfigARB || glXGetVisualFromFBConfigARB)
		return false;*/
	// get x-window extension entry points: need glXChooseFBConfig and glXGetVisualFromFBConfig functions

	int n_default_bpc_color = 8;
	int n_default_bpc_alpha = 0;
	static int p_visual_attrib_list[] = {
        GLX_X_RENDERABLE, True,
        GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
        GLX_RENDER_TYPE, GLX_RGBA_BIT,
        GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
        GLX_RED_SIZE, (n_bpp == 32)? 8 : (n_bpp == 16)? 5 : (n_bpp == 15)? 5 : n_default_bpc_color,
        GLX_GREEN_SIZE, (n_bpp == 32)? 8 : (n_bpp == 16)? 6 : (n_bpp == 15)? 5 : n_default_bpc_color,
        GLX_BLUE_SIZE, (n_bpp == 32)? 8 : (n_bpp == 16)? 5 : (n_bpp == 15)? 5 : n_default_bpc_color,
        GLX_ALPHA_SIZE, (n_bpp == 32)? 8 : (n_bpp == 16 || n_bpp == 15)? 0 : n_default_bpc_alpha,
        GLX_DEPTH_SIZE, n_depth_bpp,
        GLX_STENCIL_SIZE, n_stencil_bpp,
        GLX_DOUBLEBUFFER, True,
   		None
   	};
	// visual attributes

	m_h_dc = h_display;
	// get DC

	int n_fb_num = 0;
   	GLXFBConfig *p_fb_cfg = glXChooseFBConfigARB(h_display, DefaultScreen(h_display), p_visual_attrib_list, &n_fb_num);
    if(n_fb_num < 1 || !p_fb_cfg) {
		fprintf(stderr, "error: no suitable pixel format\n");
		return false;
	}
	m_t_pixel_format = p_fb_cfg[0];
	// choose suitable pixel format

	XVisualInfo *p_vi = glXGetVisualFromFBConfigARB(h_display, m_t_pixel_format);
	{
		Colormap t_colormap = XCreateColormap(h_display, RootWindow(h_display, p_vi->screen), p_vi->visual, AllocNone);
		XSetWindowAttributes t_swa;
	   	t_swa.colormap = t_colormap;
	    t_swa.border_pixel = 0;
		XChangeWindowAttributes(h_display, h_wnd, CWBorderPixel | CWColormap, &t_swa);
	}
	// change window colormap so OpenGL can be used with that window

	m_h_wnd = h_wnd;
	// remember window handle

	if(!(m_h_glrc = glXCreateContext(h_display, p_vi, 0, GL_TRUE))) {
		fprintf(stderr, "error: failed to create OpenGL rendering context\n");
		return false;
	}
	if(!glXMakeCurrent(h_display, h_wnd, m_h_glrc)) {
		fprintf(stderr, "error: failed to make OpenGL rendering context current\n");
		return false;
	}
	// create OpenGL context and bind it

	XFree(p_vi);
	// don't need this anymore

	if(b_forward_compatible) {
		if(!glXCreateContextAttribsARB && n_GetCreateContextARBFuncPointers()) {
			fprintf(stderr, "error: OpenGL 3.0 not supported by the X Window system\n");
			return false;
		}
		// needs glXCreateContextAttribsARB extension

		GLXContext h_old_glrc = m_h_glrc;
		// remember current (dummy) context

		const int p_params[] = {
			//WGL_CONTEXT_LAYER_PLANE_ARB, 0, // main plane // linux doesn't like this (gives GLX error)
			WGL_CONTEXT_MAJOR_VERSION_ARB, n_opengl_major,
			WGL_CONTEXT_MINOR_VERSION_ARB, n_opengl_minor,
			WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB, // want forward compatible context
			//WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB, // want core profile // nvidia drivers have problem with this
			0
		};
		if(!(m_h_glrc = glXCreateContextAttribsARB(m_h_dc, m_t_pixel_format, 0, True, p_params))) {
			fprintf(stderr, "error: Unable to create OpenGL 3.0 rendering context.\n");
			return false;
		}
		glXMakeCurrent(m_h_dc, None, NULL);
		glXDestroyContext(m_h_dc, h_old_glrc);
		if(!glXMakeCurrent(m_h_dc, h_wnd, m_h_glrc)) {
			fprintf(stderr, "error: Unable to activate the GL rendering context.\n");
			return false;
		}
		// creates forward compatible OpenGL context
	}
#endif //_WIN32 || _WIN64

    glViewport(0, 0, n_width, n_height);
    // sets viewport

    m_n_width = n_width;
    m_n_height = n_height;
    m_b_status = true;
    m_b_fullscreen = b_fullscreen;

    return true;
}

bool CGL30Driver::Shutdown()
{
	if(!m_b_status)
		return true;

    bool b_result = true;

    m_b_status = false;

#if defined(_WIN32) || defined(_WIN64)
    if(m_b_fullscreen) {
        ChangeDisplaySettings(NULL, 0);
        ShowCursor(true);
    }
	// returns from fullscreen mode

    if(m_h_glrc) {
        if(!wglMakeCurrent(NULL, NULL)) {
            MessageBoxA(NULL, "Release of DC and RC failed.", "OpenGL driver", MB_OK | MB_ICONERROR);
            b_result = false;
        }
        if(!wglDeleteContext(m_h_glrc)) {
            MessageBoxA(NULL, "Release rendering context failed.", "OpenGL driver", MB_OK | MB_ICONERROR);
            m_h_glrc = NULL;
            b_result = false;
        }
    }

    if(m_h_dc && !ReleaseDC(m_h_wnd, m_h_dc)) {
		MessageBoxA(NULL, "Release device context failed.", "OpenGL driver", MB_OK | MB_ICONERROR);
        m_h_dc = NULL;
        b_result = false;
    }
#else //_WIN32 || _WIN64
	if(!glXMakeCurrent(m_h_dc, None, NULL)) {
        fprintf(stderr, "error: Release of DC and RC failed\n");
        b_result = false;
	}
	glXDestroyContext(m_h_dc, m_h_glrc);
	m_h_glrc = NULL;
#endif //_WIN32 || _WIN64

    return b_result;
}

bool CGL30Driver::b_Status() const
{
	return m_b_status;
}

bool CGL30Driver::MakeCurrent()
{
#if defined(_WIN32) || defined(_WIN64)
    if(!wglMakeCurrent(m_h_dc, m_h_glrc))
        return false;
#else //_WIN32 || _WIN64
	if(!glXMakeCurrent(m_h_dc, m_h_wnd, m_h_glrc))
        return false;
#endif //_WIN32 || _WIN64
	return true;
}

void CGL30Driver::ResetViewport() const
{
    glViewport(0, 0, m_n_width, m_n_height);
}

void CGL30Driver::Blit() const
{
#if defined(_WIN32) || defined(_WIN64)
    SwapBuffers(m_h_dc);
#else //_WIN32 || _WIN64
	glXSwapBuffers(m_h_dc, m_h_wnd);
#endif //_WIN32 || _WIN64
}

/*
 *								=== ~CGL30Driver ===
 */

