/**
 *	@file Windowing.h
 *	@brief This contains some stuff that should've been in Main.cpp, but ended up here to reduce clutter.
 */

/*
 *	<__unnamed_0_>
 *		- mouse-button names
 */
enum {
	button_Left = 1,
	button_Right = 2,
	button_Middle = 4
};

LRESULT CALLBACK WndProc(HWND h_wnd, UINT n_msg, WPARAM n_w_param, LPARAM n_l_param);
void glPostRedisplay();
void glSwapBuffers();
void Exit();
bool onInitialize();
void onCleanup();
void onDisplay();
void onResize(int n_x, int n_y, int n_new_width, int n_new_height);
void onMouseClick(int n_button, bool b_pressing_button, int n_x, int n_y);
void onMouseMotion(int n_button, int n_x, int n_y);
void onPassiveMouseMotion(int n_x, int n_y);
void onKeyPressed(int n_key);
// function prototypes

#ifndef UNUSED
#define UNUSED(x) x
#endif // UNUSED
// this tells compiler not to complain about unused variable, decorated with UNUSED

static CGL30Driver ogl_drv;
static HWND h_wnd;

int WINAPI WinMain(HINSTANCE h_instance, HINSTANCE h_prev_instance, PSTR p_s_cmd_line, int n_cmd_show)
{
	WNDCLASSEX t_wnd_class;
	t_wnd_class.cbSize = sizeof(WNDCLASSEX);
	t_wnd_class.style = CS_HREDRAW | CS_VREDRAW;
	t_wnd_class.lpfnWndProc = WndProc;
	t_wnd_class.cbClsExtra = 0;
	t_wnd_class.cbWndExtra = 0;
	t_wnd_class.hInstance = h_instance;
	t_wnd_class.hIcon = LoadIcon(0, IDI_APPLICATION);
	t_wnd_class.hCursor = LoadCursor(0, IDC_ARROW);
	t_wnd_class.hbrBackground = 0;
	t_wnd_class.lpszMenuName = 0;
	t_wnd_class.lpszClassName = p_s_class_name;
	t_wnd_class.hIconSm = LoadIcon(0, IDI_APPLICATION);
	if(!RegisterClassEx(&t_wnd_class))
		return -1;
	// register window class

	int n_x, n_y;
	if(b_fullscreen) {
		n_x = 0;
		n_y = 0;
		n_width = GetSystemMetrics(SM_CXSCREEN);
		n_height = GetSystemMetrics(SM_CYSCREEN);
	} else if(b_centered) {
		if(n_width == CW_USEDEFAULT)
			n_width = 640;
		if(n_height == CW_USEDEFAULT)
			n_height = 480;
		n_x = (GetSystemMetrics(SM_CXSCREEN) - n_width) / 2;
		n_y = (GetSystemMetrics(SM_CYSCREEN) - n_height) / 2;
	} else {
		n_x = CW_USEDEFAULT;
		n_y = CW_USEDEFAULT;
	}
	// position the window

	RECT t_rect;
	t_rect.left = 0;
	t_rect.top = 0;
	t_rect.right = n_width;
	t_rect.bottom = n_height;
	AdjustWindowRectEx(&t_rect, (b_window_frame)? WS_OVERLAPPEDWINDOW : WS_POPUP,
		FALSE, WS_EX_APPWINDOW | ((b_window_frame)? WS_EX_WINDOWEDGE : 0));
	// adjust window rect, based on window style

	if(!(h_wnd = CreateWindowEx((b_topmost)? WS_EX_TOPMOST : 0, p_s_class_name, p_s_window_name,
	   (b_window_frame)? WS_OVERLAPPEDWINDOW : WS_POPUP,
	   n_x, n_y, t_rect.right - t_rect.left, t_rect.bottom - t_rect.top,
	   NULL, NULL, h_instance, NULL)))
		return -1;
	// create window

	if(!ogl_drv.Init(h_wnd, true, 3, 3, n_width, n_height, 32, 24, 0, false)) {
		if(!ogl_drv.Init(h_wnd, false, 0, 0, n_width, n_height, 32, 24, 0, false)) {
			fprintf(stderr, "OpenGL initialization failed\n");
			return -1;
		}
		fprintf(stderr, "running in OpenGL 2.0 compatible mode\n");
		// fallback to a non-forward-compatible OpenGL mode
	}
	// init open-gl (try OpenGL 3.3 forward-compatible profile)

	if(CGLES20ExtensionHandler::n_GetAllSupportedExtensionsFuncPointers())
		fprintf(stderr, "warning: incomplete OpenGL ES support!\n");
	// check OpenGL extensions

	if(!onInitialize())
		return -1;
	// initialize

	ShowWindow(h_wnd, n_cmd_show);
	UpdateWindow(h_wnd);
	// show window (first frame is drawn)

	if(!b_show_cursor)
		ShowCursor(false);
	// hide mouse cursor

	MSG t_msg;
	while(GetMessage(&t_msg, 0, 0, 0) > 0) { // don't use h_wnd here, it causes the window to fail to repaint in windows 7
		TranslateMessage(&t_msg);
		DispatchMessage(&t_msg);
	}
	// message loop

	if(!b_show_cursor)
		ShowCursor(true);
	// show mouse cursor

	// do not place cleanup code here, put it inside onCleanup() instead

	return 0;
}

int main(int n_arg_num, const char **p_arg_list)
{
	WinMain(GetModuleHandle(0), 0, (PSTR)p_arg_list[0], SW_SHOW);
	// i'm an idiot and i've created console application project; just call the other entry point

	return 0;
}

LRESULT CALLBACK WndProc(HWND h_wnd, UINT n_msg, WPARAM n_w_param, LPARAM n_l_param)
{
	switch(n_msg) {
	case WM_PAINT:
		{
			PAINTSTRUCT t_ps;
			HDC UNUSED(h_dc) = BeginPaint(h_wnd, &t_ps);
			EndPaint(h_wnd, &t_ps);
		}
		onDisplay();
		return 0;

	case WM_SIZE:
		{
			RECT t_wnd_area;
			GetClientRect(h_wnd, &t_wnd_area);
			int n_ww = t_wnd_area.right - t_wnd_area.left;
			int n_wh = t_wnd_area.bottom - t_wnd_area.top;
			onResize(0, 0, n_ww, n_wh);
		}
		return 0;

	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
		{
			RECT t_wnd_area;
			GetClientRect(h_wnd, &t_wnd_area);
			int n_wh = t_wnd_area.bottom - t_wnd_area.top;
			// get current window height

			onMouseClick(button_Left, n_msg == WM_LBUTTONDOWN,
				LOWORD(n_l_param), n_wh - 1 - HIWORD(n_l_param));
		}
		return 0;
	case WM_RBUTTONUP:
	case WM_RBUTTONDOWN:
		{
			RECT t_wnd_area;
			GetClientRect(h_wnd, &t_wnd_area);
			int n_wh = t_wnd_area.bottom - t_wnd_area.top;
			// get current window height

			onMouseClick(button_Right, n_msg == WM_RBUTTONDOWN,
				LOWORD(n_l_param), n_wh - 1 - HIWORD(n_l_param));
		}
		return 0;
	case WM_MBUTTONUP:
	case WM_MBUTTONDOWN:
		{
			RECT t_wnd_area;
			GetClientRect(h_wnd, &t_wnd_area);
			int n_wh = t_wnd_area.bottom - t_wnd_area.top;
			// get current window height

			onMouseClick(button_Middle, n_msg == WM_MBUTTONDOWN,
				LOWORD(n_l_param), n_wh - 1 - HIWORD(n_l_param));
		}
		return 0;
	case WM_MOUSEMOVE:
		{
			RECT t_wnd_area;
			GetClientRect(h_wnd, &t_wnd_area);
			int n_wh = t_wnd_area.bottom - t_wnd_area.top;
			// get current window height

			if(n_w_param & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON)) {
				int n_button = 0;
				if(n_w_param & MK_LBUTTON)
					n_button |= button_Left;
				if(n_w_param & MK_RBUTTON)
					n_button |= button_Right;
				if(n_w_param & MK_MBUTTON)
					n_button |= button_Middle;
				// create mask of pressed buttons

				onMouseMotion(n_button, LOWORD(n_l_param), n_wh - 1 - HIWORD(n_l_param));
			} else
				onPassiveMouseMotion(LOWORD(n_l_param), n_wh - 1 - HIWORD(n_l_param));
		}
		return 0;

	case WM_KEYDOWN:
		onKeyPressed(n_w_param);
		return 0;

	case WM_DESTROY:
		onCleanup();
		// call this first !!

		ogl_drv.Shutdown();
		// shutdown OpenGL then

		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(h_wnd, n_msg, n_w_param, n_l_param);
}


/**
 *	@brief swaps OpenGL front and back buffer, effectively making back buffer contents visible
 */
void glSwapBuffers()
{
	ogl_drv.Blit();
}

/**
 *	@brief forces the window to repaint
 */
void glPostRedisplay()
{
	InvalidateRect(h_wnd, 0, false);
}

/**
 *	@brief use this function to end the application properly
 */
void Exit()
{
	DestroyWindow(h_wnd);
}
